package net.knarcraft.ffmpegconverter.converter.sorter;

import net.knarcraft.ffmpegconverter.streams.SubtitleStream;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

/**
 * A sorter for filtering subtitle streams by title
 */
public class SubtitleTitleSorter extends AbstractSorter<SubtitleStream> {

    private final String titleFilter;

    /**
     * Instantiates a new subtitle title sorter
     *
     * <p>If a simple string, or invalid RegEx is given, any stream containing the titleFilter in its title will be
     * retained. If a valid RegEx is given, any stream matching the titleFilter is retained.</p>
     *
     * @param titleFilter <p>The filter to use. RegEx match, or a string the title must contain.</p>
     */
    public SubtitleTitleSorter(@NotNull String titleFilter) {
        this.titleFilter = titleFilter;
    }

    @Override
    public @NotNull List<SubtitleStream> sort(@NotNull List<SubtitleStream> input) {
        List<SubtitleStream> output = new ArrayList<>(input);
        if (!this.titleFilter.trim().isEmpty()) {
            if (isValidRegularExpression(this.titleFilter) && hasSpecialRegexCharacters(this.titleFilter)) {
                output.removeIf((stream) -> !stream.getTitle().matches(this.titleFilter));
            } else {
                output.removeIf((stream) -> !stream.getTitle().contains(this.titleFilter));
            }
        }
        return output;
    }

    /**
     * Checks whether the given string is a valid regular expression
     *
     * @param input <p>The string to check</p>
     * @return <p>True if the given string has no invalid expressions</p>
     */
    private static boolean isValidRegularExpression(@NotNull String input) {
        try {
            Pattern.compile(input);
            return true;
        } catch (PatternSyntaxException e) {
            return false;
        }
    }

    /**
     * Checks whether the input string has any RegEx special characters
     *
     * @param input <p>The input to check</p>
     * @return <p>True if RegEx characters exist in the string</p>
     */
    private static boolean hasSpecialRegexCharacters(String input) {
        Pattern regexSpecialCharacters = Pattern.compile("[\\\\.\\[\\]{}()<>*+\\-=!?^$|]");
        return regexSpecialCharacters.matcher(input).find();
    }

}
